# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from oslo_utils import uuidutils

from networking_fujitsu.ml2.fossw import mech_fossw
from neutron.plugins.ml2 import config as ml2_config
from neutron.tests.unit.plugins.ml2 import test_plugin as test_ml2_plugin

LLI = {'single': [{"switch_id": "00:00:4c:ee:e5:39", " port_id": "0/1",
                   "switch_info": "ET-7648BRA-FOS"}],
       'lag': [{"switch_id": "00:00:4c:ee:e5:39", "port_id": "0/2",
                "switch_info": "ET-7648BRA-FOS"},
               {"switch_id": "00:00:4c:ee:e5:39", "port_id": "0/3",
                "switch_info": "ET-764BRA-FOS"}],
       'mlag': [{"switch_id": "00:00:4c:ee:e5:39", "port_id": "0/2",
                 "switch_info": "ET-764BRA-FOS"},
                {"switch_id": "00:00:4c:ee:e5:40", "port_id": "0/2",
                 "switch_info": "ET-764BRA-FOS"}]}


class TestFujitsuMechDriverV2(test_ml2_plugin.Ml2PluginV2TestCase):
    """Test Fujitsu mechanism driver.

    """

    _mechanism_drivers = ['fujitsu_fossw']

    _skip = ["test_create_router_port_and_fail_create_postcommit"]

    def setup(self):

        ml2_config.cfg.CONF.set_override(
            'tenant_network_types', ['vlan', 'vxlan'], 'ml2')

        if self._testMethodName in self._skip:
            self.skipTest("This test has already verified at neutron's test.")

        def mocked_initialize(self):
            self._list_switch_info = [{"ip": "192.168.3.123", "port": 22,
                                       "username": "admin", "password": "",
                                       "timeout": 30},
                                      {"ip": "192.168.3.124", "port": 22,
                                       "username": "admin", "password": "",
                                       "timeout": 30}]
            self._driver = mock.MagicMock()

            with mock.patch.object(mech_fossw.FujitsuMechanism,
                                   'initialize', new=mocked_initialize):
                super(TestFujitsuMechDriverV2, self).setUp()


class TestFujitsuMechDriverNetworksV2(test_ml2_plugin.TestMl2NetworksV2,
                                      TestFujitsuMechDriverV2):
    pass


class TestFujitsuMechDriverPortsV2(test_ml2_plugin.TestMl2PortsV2,
                                   TestFujitsuMechDriverV2):
    pass


class TestFujitsuMechDriverBaremetalPortsV2(TestFujitsuMechDriverV2):

    def setup_net_and_port(self, method='create', net={}, port={},
                           phy_port='single'):
        def call_method(arg):
            return 'new_' + arg + '_request'
        target = call_method(method)

        with self.network() as network:
            net_id = network['network']['id']
            self._create_subnet(self.fmt, net_id, '172.16.1.0/24')
            port_data = {'port': {'network_id': network['network']['id'],
                                  'tenant_id':
                                  network['network']['tenant_id'],
                                  'name': 'prov-port', 'admin_state_up': 1,
                                  'binding:vnic_type': 'baremetal',
                                  'binding:profile':
                                  {'local_link_information':
                                      LLI[phy_port]}}}
            if method in ['update', 'delete']:
                port = self.deserialize(
                    self.fmt,
                    self.new_create_request(
                        'ports',
                        port_data
                    ).get_response(self.api)
                )
                p_id = port['port']['id']
                call_target = getattr(self, target)
                if method is 'update':
                    body = {
                        'port': {
                            'binding:host_id': uuidutils.generate_uuid()
                        }
                    }
                    call_target('ports', p_id,
                                str(body)).get_response(self.api)
                else:
                    call_target('ports', p_id).get_response(self.api)
            else:
                call_target = getattr(self, target)
                call_target('ports', port_data).get_response(self.api)

    def test_create_with_single_NIC(self):
        self.setup_net_and_port()

    def test_create_with_lag(self):
        self.setup_net_and_port(phy_port='lag')

    def test_create_with_mlag(self):
        self.setup_net_and_port(phy_port='mlag')

    def test_update_with_single_NIC(self):
        self.setup_net_and_port(method='update')

    def test_update_with_lag(self):
        self.setup_net_and_port(method='update', phy_port='lag')

    def test_update_with_mlag(self):
        self.setup_net_and_port(method='update', phy_port='mlag')

    def test_delete_with_single_NIC(self):
        self.setup_net_and_port(method='delete')

    def test_delete_with_lag(self):
        self.setup_net_and_port(method='delete', phy_port='lag')

    def test_delete_with_mlag(self):
        self.setup_net_and_port(method='delete', phy_port='mlag')



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Fujitsu fossw Driver implements CLI over ssh for Neutron network life-cycle
management.
"""

from oslo_config import cfg
from oslo_log import log as logging

from networking_fujitsu._i18n import _LE
from networking_fujitsu._i18n import _LW
from networking_fujitsu.ml2.fossw import client
from neutron.common import utils

LOG = logging.getLogger(__name__)
_LOCK_NAME = 'fujitsu-fossw'


class FOSSWVlanDriver(object):
    """Fossw CLI interface driver for Neutron network.

    Handles life-cycle management of Neutron network.
    """

    def __init__(self, conf=None):
        if conf:
            self._conf = conf
        else:
            self._conf = cfg.CONF
        self.client = client.FOSSWClient(self._conf)

    @utils.synchronized(_LOCK_NAME)
    def get_switch_mac_ip_pair(self, ips):
        """Get and create pair of switch's MAC address and IP address.

        :param ips: list of IP address of FOS switch
        :type ips: list of IP address strings

        :returns: the dictionary which stores pair of  MAC address and IP
                  address
        :rtype: dictionary of MAC address and IP address

        """
        switches_mac_ip_pair = {}
        for ip in ips:
            self.client.connect(ip)
            res_mac = self.client.get_switch_mac()
            index = res_mac.rfind('. ') + 2
            switch_mac = res_mac[index:(index + 17)]
            switches_mac_ip_pair.update({switch_mac.lower(): ip})
            self.client.disconnect()
        return switches_mac_ip_pair

    def create_vlan(self, ip, vlan_id):
        """Create VLAN with specified VLAN ID.

        :param ip: IP address of target FOS switch
        :type ip: string
        :param vlan_id: ID of VLAN to be created
        :type vlan_id: string

        :returns: None
        :rtype: None

        """
        method = 'create_vlan'
        try:
            self.client.connect(ip)
            self.client.create_vlan(vlan_id)
            self.client.disconnect()
        except Exception as e:
            self.client.disconnect()
            LOG.exception(_LE("an error occurred while creating vlan to FOS "
                              "switch. %s"), e)
            raise client.FOSSWClientException(method)

    def delete_vlan(self, ip, vlan_id):
        """Delete VLAN with specified VLAN ID.

        :param ip: IP address of target FOS switch
        :type ip: string
        :param vlan_id: ID of VLAN to be deleted
        :type vlan_id: string

        :returns: None
        :rtype: None

        """
        method = 'delete_vlan'
        try:
            self.client.connect(ip)
            self.client.delete_vlan(vlan_id)
            self.client.disconnect()
        except Exception as e:
            self.client.disconnect()
            LOG.exception(_LE("An error occurred while deleting vlan from "
                              "FOS switch. %s"), e)
            raise client.FOSSWClientException(method)

    def setup_vlan(self, vlan_id, lli, ip_mac_pairs):
        """Setup VLAN for physical port on FOS Switch.

        :param vlan_id: the ID of VLAN to be associated
        :type vlan_id: string
        :param lli: the local link information of ironic node
        :type lli: list of local link information dictionary
        :param ip_mac_pairs: the pair of MAC address and IP address of FOS
                             switch
        :type ip_mac_pairs: dictionary

        :returns: None
        :rtype: None

        """
        method = "setup_vlan"
        sw_mac = lli[0]['switch_id']
        port_id = lli[0]['port_id']

        try:
            target_ip = ip_mac_pairs[sw_mac]
        except KeyError as e:
            LOG.exception(_LE("FOS switch with MAC(%s) not found."), sw_mac)
            raise client.FOSSWClientException(method)
        try:
            self.client.connect(target_ip)
            self.client.set_vlan(vlan_id, port_id)
            self.client.disconnect()
        except Exception as e:
            self.client.disconnect()
            LOG.exception(_LE("An error occurred while setup vlan for "
                              "physical port on FOS switch. %s"), e)
            raise client.FOSSWClientException(method)

    @utils.synchronized(_LOCK_NAME, external=True)
    def setup_vlan_with_lag(self, vlan_id, llis, ip_mac_pairs):
        """Setup VLAN and LAG for physical ports FOS switch.

        :param vlan_id: the ID of VLAN to be associated
        :type vlan_id: string
        :param llis: the local link informations of ironic node
        :type llis: list of local link informations dictionary
        :param ip_mac_pairs: the pair of MAC address and IP address of FOS
                             switch
        :type ip_mac_pairs: dictionary

        :returns: None
        :rtype: None

        """
        method = "setup_vlan_with_lag"
        mlag = None
        switch_mac_list = []
        for lli in llis:
            switch_mac_list.append(lli['switch_id'])
        unique_mac_list = set(switch_mac_list)
        if len(unique_mac_list) > 1:
            mlag = self.is_valid_mlag(unique_mac_list, ip_mac_pairs)
            if not mlag:
                LOG.exception(
                    _LE("Specified switches cannot set mLAG pair. "
                        "Please confirm each switch's peerlink setting.")
                )
                raise client.FOSSWClientException(method)

        # setup vlan for each physical port
        for lli in llis:
            arr_lli = []
            arr_lli.append(lli)
            self.setup_vlan(vlan_id, arr_lli, ip_mac_pairs)

        # Create lag resource
        lag = {}
        for mac in unique_mac_list:
            target_ip = ip_mac_pairs[mac]

            lag_ports = []
            for lli in llis:
                if lli['switch_id'] == mac:
                    lag_ports.append(lli['port_id'])

            lag.update({mac: lag_ports})
            self.client.connect(target_ip)
            lag_portname = self.client.get_free_logical_port()
            for port in lag[mac]:
                self.client.join_to_lag(port, lag_portname)
            if mlag:
                # Get free VPC id from fossw.
                vpcid = self.client.get_vpcid()
                if vpcid:
                    self.client.join_to_vpc(target_ip, lag_portname, vpcid)
                else:
                    # All vpc is already used by other logical ports.
                    self.client.disconnect()
                    raise client.FOSSWClientException(method)
            self.client.disconnect()

    def is_valid_mlag(self, macs, ip_mac_pairs):
        """Validates which given pair of MAC address is valid mlag pair or not.

        :param macs: list of MAC address which mlag candidacy pair
        :type macs: list
        :param ip_mac_pairs: the pair of MAC address and IP address of FOS
                             switches
        :type ip_mac_pairs: dictionary

        :returns: Result of validation
        :rtype: Boolean

        """
        method = "is_valid_mlag"
        try:
            first_switch_ip = ip_mac_pairs[macs[0]]
            self.client.connect(first_switch_ip)
            mlag_partner_ip = self.client.get_peerlink_partner()
            self.client.disconnect()
            if mlag_partner_ip == macs[1]:
                return True
            else:
                return False
        except Exception as e:
            self.client.disconnect()
            LOG.exception(_LE("An error occurred while validating specified "
                              "FOS switches are VPC pair. %s"), e)
            raise client.FOSSWClientException(method)

    def clear_vlan(self, vlan_id, lli, ip_mac_pairs):
        """Clear VLAN from FOS switch.

        :param vlan_id: the ID of VLAN to be disassociated
        :type vlan_id: string
        :param lli: the local link information of ironic node
        :type lli: list
        :param ip_mac_pairs: the pair of MAC address and IP address of FOS
                             switches
        :type ip_mac_pairs: dictionary
        """
        method = "clear_vlan"
        sw_mac = lli[0]['switch_id']
        port_id = lli[0]['port_id']

        try:
            target_ip = ip_mac_pairs[sw_mac]
        except KeyError as e:
            LOG.exception(_LE("FOS switch with MAC(%s) not found."), sw_mac)
            raise client.FOSSWClientException(method)
        try:
            self.client.connect(target_ip)
            self.client.clear_vlan(vlan_id, port_id)
            self.client.disconnect()
        except Exception as e:
            self.client.disconnect()
            LOG.exception(_LE("an error occurred while clearing VLAN from "
                              "physical port on FOS switch. %s"), e)
            raise client.FOSSWClientException(method)

    @utils.synchronized(_LOCK_NAME, external=True)
    def clear_vlan_with_lag(self, vlan_id, llis, ip_mac_pairs):
        """Clear VLAN and LAG from FOS switch.

        :param vlan_id: the ID of VLAN to be disassociated
        :type vlan_id: string
        :param llis: the local link informations of ironic node
        :type vlan_id: list
        :param ip_mac_pairs: the pair of MAC address and IP address of FOS
                             switches
        :type ip_mac_pairs: dictionary

        :returns: None
        :rtype: None

        """
        method = "clear_vlan_with_lag"
        mlag = None
        switch_mac_list = []
        for lli in llis:
            switch_mac_list.append(lli['switch_id'])
        unique_mac_list = set(switch_mac_list)
        if len(unique_mac_list) > 1:
            mlag = self.is_valid_mlag(unique_mac_list, ip_mac_pairs)
            if not mlag:
                LOG.exception(_LE("Specified switches are not able to clear "
                                  "mLAG pair. Please confirm each switch's "
                                  "peerlink setting."))
                raise client.FOSSWClientException(method)

        for mac in unique_mac_list:
            target_ip = ip_mac_pairs[mac]
            ports = []
            for lli in llis:
                if lli['switch_id'] == mac:
                    ports.append(lli['port_id'])
            self.client.connect(target_ip)
            lag_portname = self.client.get_lag_port(ports[0])

            if mlag:
                vpcid = self.client.get_vpcid(lag_portname)
                if vpcid:
                    self.client.leave_from_vpc(lag_portname, vpcid)
                else:
                    LOG.warning(_LW("Specified logical port is not "
                                    "associated with any VPC on the FOSSW. "
                                    "skip leave from vpc."))
            for port in ports:
                self.client.leave_from_lag(port, lag_portname)
            self.client.disconnect()

        # Clear lag setting
        for lli in llis:
            arr_lli = []
            arr_lli.append(lli)
            self.clear_vlan(vlan_id, arr_lli, ip_mac_pairs)



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import mock
from networking_fujitsu.ml2.fossw import client
from networking_fujitsu.ml2.fossw import fossw_vlandriver
from networking_fujitsu.ml2.fossw import mech_fossw
from neutron.tests import base
from oslo_config import cfg

LLI = {'single': [{"switch_id": "00:00:4c:ee:e5:39", "port_id": "0/1",
                   "switch_info": "ET-7648BRA-FOS"}],
       'lag': [{"switch_id": "00:00:4c:ee:e5:39", "port_id": "0/1",
                "switch_info": "ET-7648BRA-FOS"},
               {"switch_id": "00:00:4c:ee:e5:39", "port_id": "0/2",
                "switch_info": "ET-7648BRA-FOS"}],
       'mlag': [{"switch_id": "00:00:4c:ee:e5:39", "port_id": "0/1",
                 "switch_info ": "ET-7648BRA-FOS"},
                {"switch_id": "00:00:4c:ee:e5:40", "port_id": "0/1",
                 "switch_info": "ET-7648BRA-FOS"}]}

DUMMY_FOSSW_IPS = ['192.168.0.1']


class BaseTestFOSSWVlanDriver(base.BaseTestCase):
    """Base class to test Fujitsu FOSSW driver."""

    def setUp(self):
        super(BaseTestFOSSWVlanDriver, self).setUp()
        cfg.CONF.register_opts(
            mech_fossw.ML2_FUJITSU,
            mech_fossw.ML2_FUJITSU_GROUP
        )
        self.driver = fossw_vlandriver.FOSSWVlanDriver(cfg.CONF)


class TestFOSSWVlanDriverCreateVlan(BaseTestFOSSWVlanDriver):
    """Test Fujitsu fossw driver for create VLAN"""

    def setUp(self):
        super(TestFOSSWVlanDriverCreateVlan, self).setUp()
        self.driver.client.connect = mock.Mock(return_value=None)
        self.driver.client.create_vlan = mock.Mock(return_value=None)
        self.driver.client.disconnect = mock.Mock(return_value=None)

    def test_create_vlan(self):
        self.driver.create_vlan(DUMMY_FOSSW_IPS[0], 2)
        self.assertIsNone(self.driver.client.ssh)

    def test_create_vlan_fail(self):
        error = client.FOSSWClientException('create_vlan')
        self.driver.client.connect.side_effect = error
        self.assertRaises(
            client.FOSSWClientException,
            self.driver.create_vlan,
            DUMMY_FOSSW_IPS[0],
            2
        )


class TestFOSSWVlanDriverDeleteVlan(BaseTestFOSSWVlanDriver):
    """Test Fujitsu FOSSW vlan driver for delete VLAN"""
    def setUp(self):
        super(TestFOSSWVlanDriverDeleteVlan, self).setUp()
        self.driver.client.connect = mock.Mock(return_value=None)
        self.driver.client.delete_vlan = mock.Mock(return_value=None)
        self.driver.client.disconnect = mock.Mock(return_value=None)

    def test_delete_vlan(self):
        self.driver.delete_vlan(DUMMY_FOSSW_IPS[0], 2)
        self.assertIsNone(self.driver.client.ssh)

    def test_delete_vlan_fail(self):
        error = client.FOSSWClientException('delete_vlan')
        self.driver.client.connect.side_effect = error
        self.assertRaises(
            client.FOSSWClientException,
            self.driver.delete_vlan,
            DUMMY_FOSSW_IPS[0],
            2
        )


class TestFOSSWVlanDriverSetupVlan(BaseTestFOSSWVlanDriver):
    """Test Fujitsu FOSSW vlan driver."""
    def setUp(self):
        super(TestFOSSWVlanDriverSetupVlan, self).setUp()
        self.driver.client.connect = mock.Mock(return_value=None)
        self.driver.client.set_vlan = mock.Mock(return_value=None)
        self.driver.client.disconnect = mock.Mock(return_value=None)

    def test_setup_vlan(self):
        ip_mac_pairs = {'00:00:4c:ee:e5:39': '192.168.1.1'}
        self.driver.setup_vlan(2, LLI['single'], ip_mac_pairs)
        self.assertIsNone(self.driver.client.ssh)

    def test_setup_vlan_fail_with_keyerror(self):
        ip_mac_pairs = {'00:00:4c:ee:e5:40': '192.168.1.1'}
        self.assertRaises(
            client.FOSSWClientException,
            self.driver.setup_vlan,
            2,
            LLI['single'],
            ip_mac_pairs
        )
        self.driver.client.disconnect.assert_not_called()

    def test_setup_vlan_fail_with_connect_fail(self):
        ip_mac_pairs = {'00:00:4c:ee:e5:39': '192.168.1.1'}
        error = client.FOSSWClientException('set_vlan')
        self.driver.client.connect.side_effect = error
        self.assertRaises(
            client.FOSSWClientException,
            self.driver.setup_vlan,
            2,
            LLI['single'],
            ip_mac_pairs
        )
        self.driver.client.set_vlan.assert_not_called()
        self.driver.client.disconnect.assert_called_once()


class TestFOSSWVlanDriverSetupVlanWithLag(BaseTestFOSSWVlanDriver):
    """Test FujitsuFOSSW vlan driver for setup vlan with lag"""
    def setUp(self):
        super(TestFOSSWVlanDriverSetupVlanWithLag, self).setUp()
        self.driver.is_valid_mlag = mock.Mock(return_value=None)
        self.driver.setup_vlan = mock.Mock(return_value=None)
        self.driver.client.connect = mock.Mock(return_value=None)
        self.driver.client.setup_vlan = mock.Mock(return_value=None)
        self.driver.client.get_free_logical_port = mock.Mock()
        self.driver.client.join_to_lag = mock.Mock(return_value=None)
        self.driver.client.disconnect = mock.Mock(return_value=None)

    def test_setup_vlan_with_lag(self):
        llis = LLI['lag']
        vlan_id = 2
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        self.driver.client.get_free_logical_port.return_value = "3/1"
        self.assertIsNone(self.driver.setup_vlan_with_lag(vlan_id,
                                                          llis,
                                                          ip_mac_pairs))
        self.driver.is_valid_mlag.assert_not_called()

    def test_setup_vlan_with_mlag(self):
        llis = LLI['mlag']
        vlan_id = 2
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        self.driver.is_valid_mlag.return_value = True
        self.driver.client.get_free_logical_port.return_value = "3/1"
        self.driver.client.get_vpcid = mock.Mock(return_value=1)
        self.driver.client.join_to_vpc = mock.Mock(return_value=None)
        self.assertIsNone(self.driver.setup_vlan_with_lag(vlan_id,
                                                          llis,
                                                          ip_mac_pairs))
        self.driver.is_valid_mlag.assert_called_once()
        self.assertEqual(2, self.driver.client.join_to_vpc.call_count)

    def test_setup_vlan_with_mlag_not_valid_pair(self):
        llis = LLI['mlag']
        vlan_id = 2
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        self.driver.is_valid_mlag.return_value = False
        self.assertRaises(client.FOSSWClientException,
                          self.driver.setup_vlan_with_lag,
                          vlan_id,
                          llis,
                          ip_mac_pairs)
        self.driver.client.setup_vlan.assert_not_called()

    def test_setup_vlan_with_mlag_vpcid_already_used(self):
        llis = LLI['mlag']
        vlan_id = 2
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        self.driver.is_valid_mlag.return_value = True
        self.driver.client.get_vpcid = mock.Mock(return_value=None)
        self.assertRaises(client.FOSSWClientException,
                          self.driver.setup_vlan_with_lag,
                          vlan_id,
                          llis,
                          ip_mac_pairs)
        self.driver.client.disconnect.assert_called_once()


class TestFOSSWVlanDriverClearVlan(BaseTestFOSSWVlanDriver):
    """Test Fujitsu FOSSW vlan driver for clear vlan"""
    def setUp(self):
        super(TestFOSSWVlanDriverClearVlan, self).setUp()
        self.driver.client.connect = mock.Mock(return_value=None)
        self.driver.client.clear_vlan = mock.Mock(return_value=None)
        self.driver.client.disconnect = mock.Mock(return_value=None)

    def test_clear_vlan(self):
        ip_mac_pairs = {'00:00:4c:ee:e5:39': '192.168.1.1'}
        self.driver.clear_vlan(2, LLI['single'], ip_mac_pairs)
        self.assertIsNone(self.driver.client.ssh)

    def test_clear_vlan_fail_with_keyerror(self):
        ip_mac_pairs = {'00:00:4c:ee:e5:40': '192.168.1.1'}
        self.assertRaises(
            client.FOSSWClientException,
            self.driver.clear_vlan,
            2,
            LLI['single'],
            ip_mac_pairs
        )
        self.driver.client.disconnect.assert_not_called()

    def test_clear_vlan_fail_with_connect_fail(self):
        ip_mac_pairs = {'00:00:4c:ee:e5:39': '192.168.1.1'}
        error = client.FOSSWClientException('clear_vlan')
        self.driver.client.connect.side_effect = error
        self.assertRaises(
            client.FOSSWClientException,
            self.driver.clear_vlan,
            2,
            LLI['single'],
            ip_mac_pairs
        )
        self.driver.client.clear_vlan.assert_not_called()
        self.driver.client.disconnect.assert_called_once()


class TestFOSSWVlanDriverClearVlanWithLAG(BaseTestFOSSWVlanDriver):
    """Test Fujitsu FOSSW vlan driver for clear vlan with LAG"""
    def setUp(self):
        super(TestFOSSWVlanDriverClearVlanWithLAG, self).setUp()
        self.driver.is_valid_mlag = mock.Mock(return_value=None)
        self.driver.client.connect = mock.Mock(return_value=None)
        self.driver.clear_vlan = mock.Mock(return_value=None)
        self.driver.client.disconnect = mock.Mock(return_value=None)
        self.driver.client.get_lag_port = mock.Mock()
        self.driver.client.get_vpcid = mock.Mock()
        self.driver.client.leave_from_lag = mock.Mock(return_value=None)
        self.driver.client.leave_from_vpc = mock.Mock(return_value=None)

    def test_clear_vlan_with_lag(self):
        llis = LLI['lag']
        vlan_id = 2
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        self.driver.client.get_lag_port.return_value = "3/1"
        self.assertIsNone(self.driver.clear_vlan_with_lag(vlan_id,
                                                          llis,
                                                          ip_mac_pairs))
        self.driver.is_valid_mlag.assert_not_called()
        self.driver.client.get_vpcid.assert_not_called()

    def test_clear_vlan_with_mlag(self):
        llis = LLI['mlag']
        vlan_id = 2
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        self.driver.is_valid_mlag.return_value = True
        self.driver.client.get_lag_port.return_value = "3/1"
        self.driver.client.get_vpcid.return_value = "1"
        self.assertIsNone(self.driver.clear_vlan_with_lag(vlan_id,
                                                          llis,
                                                          ip_mac_pairs))
        self.driver.is_valid_mlag.assert_called_once()
        self.driver.client.leave_from_vpc.assert_called_with("3/1", "1")
        self.assertEqual(2, self.driver.client.leave_from_vpc.call_count)

    def test_clear_vlan_mlag_not_defined(self):
        llis = LLI['mlag']
        vlan_id = 2
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        self.driver.is_valid_mlag.return_value = False
        self.assertRaises(client.FOSSWClientException,
                          self.driver.clear_vlan_with_lag,
                          vlan_id,
                          llis,
                          ip_mac_pairs)

    def test_clear_vlan_with_mlag_not_valid_pair(self):
        llis = LLI['mlag']
        vlan_id = 2
        ip_mac_pairs = {"00:00:4c:ee:e5:39": "192.168.1.1",
                        "00:00:4c:ee:e5:40": "192.168.1.2"}
        self.driver.is_valid_mlag.return_value = True
        self.driver.client.get_lag_port.return_value = "3/1"
        self.driver.client.get_vpcid.return_value = None
        self.assertIsNone(self.driver.clear_vlan_with_lag(vlan_id,
                                                          llis,
                                                          ip_mac_pairs))
        self.assertEqual(2, self.driver.client.disconnect.call_count)



# Copyright 2015-2016 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import os

from oslo_utils import uuidutils

from networking_fujitsu.ml2.cfab import cfabdriver
from networking_fujitsu.ml2.cfab import mech_cfab
from neutron.plugins.ml2 import config as ml2_config
from neutron.tests.unit.plugins.ml2 import test_plugin as test_ml2_plugin

LLI = {
    'single': [{
        "switch_id": "00:00:4c:ee:e5:39",
        "port_id": "1/1/0/1",
        "switch_info": "CFX2000R"}],
    'lag': [{
        "switch_id": "00:00:4c:ee:e5:39",
        "port_id": "1/1/0/2",
        "switch_info": "CFX2000R"}, {
        "switch_id": "00:00:4c:ee:e5:39",
        "port_id": "1/1/0/3",
        "switch_info": "CFX2000R"}],
    'mlag': [{
        "switch_id": "00:00:4c:ee:e5:39",
        "port_id": "1/1/0/2",
        "switch_info": "CFX2000R"}, {
        "switch_id": "00:00:4c:ee:e5:40",
        "port_id": "1/1/0/2",
        "switch_info": "CFX2000R"}]
}


class TestFujitsuMechDriverV2(test_ml2_plugin.Ml2PluginV2TestCase):
    """Test Fujitsu mechanism driver.

    If the environment variable OS_FUJITSU_CFAB_ADDRESS is defined, tests will
    be performed using the C-Fabric driver, and a telnet connection to the
    address OS_FUJITSU_CFAB_ADDRESS and the port OS_FUJITSU_CFAB_PORT (defaults
    to 23) will be established. The username will be taken from
    OS_FUJITSU_CFAB_USERNAME (defaults to "admin") and the password will be
    taken from OS_FUJITSU_CFAB_PASSWORD (defaults to "password").
    If the environment variable OS_FUJITSU_CFAB_ADDRESS is NOT defined, tests
    will be performed using the mock driver instead of the C-Fabric driver, and
    no real telnet connection will be used.
    """

    _mechanism_drivers = ['fujitsu_cfab']
    # test_create_router_port_and_fail_create_postcommit:
    #     This one is mocked 'fake_driver' only. So, our plugin's instance
    #     hasn't mocked and fail definitely. Therefore, skips this test.
    _skip = ["test_create_router_port_and_fail_create_postcommit"]

    def setUp(self):

        ml2_config.cfg.CONF.set_override(
            'tenant_network_types', ['vlan'], 'ml2')

        if self._testMethodName in self._skip:
            self.skipTest("This test has already verified at neutron's test.")

        address = os.environ.get('OS_FUJITSU_CFAB_ADDRESS')
        if address:
            ml2_fujitsu_opts = {
                'username': os.environ.get('OS_FUJITSU_CFAB_USERNAME') or
                "admin",
                'password': os.environ.get('OS_FUJITSU_CFAB_PASSWORD') or
                "password",
                'address': address,
                'physical_networks': ["physnet1:1", "physnet2:2"],
                'pprofile_prefix': "test-"}

            for opt, val in ml2_fujitsu_opts.items():
                ml2_config.cfg.CONF.set_override(opt, val, "fujitsu_cfab")

            try:
                cfabdriver.TELNET_PORT = int(
                    os.environ['OS_FUJITSU_CFAB_PORT'])
            except KeyError:
                pass
            super(TestFujitsuMechDriverV2, self).setUp()
        else:
            def mocked_initialize(self):
                self._switch = {'address': "", 'username': "", 'password': ""}
                self._driver = mock.MagicMock()
                self._physical_networks = {'physnet1': "1", 'physnet2': "2"}

            with mock.patch.object(mech_cfab.CFABMechanismDriver,
                                   'initialize', new=mocked_initialize):
                super(TestFujitsuMechDriverV2, self).setUp()


class TestFujitsuMechDriverPortsV2(test_ml2_plugin.TestMl2PortsV2,
                                   TestFujitsuMechDriverV2):
    pass


class TestFujitsuMechDriverBaremetalPortsV2(TestFujitsuMechDriverV2):

    def setup_net_and_port(self, method='create', net={}, port={},
                           phy_port='single'):

        def call_method(arg):
            return 'new_' + arg + '_request'
        target = call_method(method)
        with self.network() as network:
            net_id = network['network']['id']
            self._create_subnet(self.fmt, net_id, '172.16.1.0/24')
            port_data = {
                'port': {
                    'network_id': network['network']['id'],
                    'tenant_id': network['network']['tenant_id'],
                    'name': 'prov-port',
                    'admin_state_up': 1,
                    'binding:vnic_type': 'baremetal',
                    'binding:profile': {
                        'local_link_information': LLI[phy_port]
                    }
                }
            }
            if method in ['update', 'delete']:
                port = self.deserialize(
                    self.fmt, self.new_create_request(
                        'ports', port_data).get_response(self.api))
                p_id = port['port']['id']
                call_target = getattr(self, target)
                if method is 'update':
                    body = {
                        'port': {'binding:host_id': uuidutils.generate_uuid()}
                    }
                    call_target('ports', p_id, str(body)).get_response(
                        self.api)
                else:
                    call_target('ports', p_id).get_response(self.api)
            else:
                call_target = getattr(self, target)
                call_target('ports', port_data).get_response(self.api)

    def test_create_with_single_NIC(self):
        self.setup_net_and_port()

    def test_create_with_lag(self):
        self.setup_net_and_port(phy_port='lag')

    def test_create_with_mlag(self):
        self.setup_net_and_port(phy_port='mlag')

    def test_update_with_single_NIC(self):
        self.setup_net_and_port(method="update")

    def test_update_with_lag(self):
        self.setup_net_and_port(method='update', phy_port='lag')

    def test_update_with_mlag(self):
        self.setup_net_and_port(method='update', phy_port='mlag')

    def test_delete_with_single_NIC(self):
        self.setup_net_and_port(method='delete')

    def test_delete_with_lag(self):
        self.setup_net_and_port(method='delete', phy_port='lag')

    def test_delete_with_mlag(self):
        self.setup_net_and_port(method='delete', phy_port='mlag')



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import paramiko
import socket
import time

from oslo_log import log as logging

from networking_fujitsu._i18n import _
from networking_fujitsu._i18n import _LE
from networking_fujitsu._i18n import _LW

LOG = logging.getLogger(__name__)

MODE_GLOBAL = 'configure'
MODE_VLAN = 'vlan database'
MODE_INTERFACE = 'interface'
ENABLE = 'enable'
END = 'end'
READ_TIMEOUT = 5.0
MAX_LOOP = 50
MAX_VPC_ID = 63
SHOW_PC_BRIEF = 'show port-channel brief'


class FOSSWClient(object):

    def __init__(self, conf):
        self._conf = conf
        self.ssh = None
        self.console = None

    def connect(self, ip):
        """Establish ssh connection for FOS switch.

        :param ip: a ip address of FOS switch
        :type ip: string

        :returns: None
        :rtype: None

        """
        method = "connect"
        retry_count = 0
        while retry_count < 5:
            try:
                self.ssh = paramiko.SSHClient()
                self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                self.ssh.connect(
                    ip,
                    port=self._conf.fujitsu_fossw.port,
                    username=self._conf.fujitsu_fossw.username,
                    password=self._conf.fujitsu_fossw.password,
                    timeout=self._conf.fujitsu_fossw.timeout
                )
                self.console = self.ssh.invoke_shell()
                self.console.settimeout(READ_TIMEOUT)
                return
            except IOError as e:
                LOG.warning(_LW('Could not initialize SSH client. %s'), e)
            except (paramiko.ssh_exception.BadHostKeyException,
                    paramiko.ssh_exception.AuthenticationException,
                    paramiko.ssh_exception.SSHException) as e:
                LOG.warning(_LW('Could not connect to FOS switch. An error'
                                'occurred while connecting. %s'), e)
            except socket.error as e:
                e_no, e_str = e
                LOG.warning(_LW('A socket error occurred while connecting.\n'
                                '[Errno %(e_no)s] %(e_str)s'),
                            {'e_no': e_no, 'e_str': e_str})
            except Exception as e:
                LOG.warning(_LW('Unexpected error occurred while connecting. '
                                '%s'), e)
            retry_count += 1
            self.disconnect()
            LOG.warning(_LW('Connect attempt %s failed.'), retry_count)
        if retry_count >= 5:
            self.disconnect()
            LOG.exception(_LE('Max retries exceeded. Failed to connect to '
                              'FOS switch.'))
            raise FOSSWClientException(method)

    def disconnect(self):
        """Close ssh connection with FOS switch.

        :returns: None
        :rtypes: None

        """
        if self.ssh:
            self.ssh.close()
            self.ssh = None
            self.console = None
        else:
            LOG.debug(_("SSH connection already closed."))

    def _exec_command(self, command):
        try:
            raw_res = ""
            self.console.send(command + "\n")
            LOG.debug(_("FOSSW client sending command: %s"), command)
            i = 0
            while i < MAX_LOOP:
                time.sleep(0.1)
                if self.console.recv_ready():
                    raw_res += self.console.recv(1024)
                elif raw_res:
                    break
                i += 1
            if (i == MAX_LOOP):
                LOG.error(_LE("No reply from FOS switch."))
                raise socket.timeout
            res = raw_res.replace('\r\n', '')
            received = res[(res.find(command) + len(command)):]
        except socket.timeout:
            self.disconnect()
            LOG.exception(_LE('Socket timeout occured while executing '
                              'commands to FOS Switch.'))
            raise FOSSWClientException('_exec_command')
        else:
            LOG.debug(_("FOSSW client received: %s"), received)
            # NOTE(yushiro) Validate received message here
            return received

    def _format_command(self, command, **kwargs):
        method = "_format_command"
        if not all(kwargs.values()):
            raise FOSSWClientException(method)
        try:
            formatted = command.format(**kwargs)
        except (KeyError, TypeError):
            raise FOSSWClientException(method)
        return formatted

    def create_vlan(self, segmentation_id):
        """Define VLAN with specified VLAN ID to FOS switch.

        :param segmentation_id: id of VLAN to be created.
        :type segmentation_id: string

        :returns: received message from FOS switch
        :rtype: string

        """
        self.change_mode(MODE_VLAN)
        cmd = self._format_command("vlan {vlan_id}", vlan_id=segmentation_id)
        return self._exec_command(cmd)

    def delete_vlan(self, segmentation_id):
        """Undefine VLAN with specified VLAN ID from FOS switch.

        :param segmentation_id: id of VLAN to be deleted.
        :type segmentation_id: string

        :returns: None
        :rtype: None

        """
        self.change_mode(MODE_VLAN)
        cmd = self._format_command("no vlan {vlan_id}",
                                   vlan_id=segmentation_id)
        if "Failed to delete" in self._exec_command(cmd):
            LOG.warning(_LW("VLAN(%s) has already deleted."), segmentation_id)

    def set_vlan(self, segmentation_id, port_id):
        """Associate VLAN to specified physical port on FOS switch.

        :param segmentation_id: id of VLAN to be associated
        :type segmentation_id: string
        :param port_id: the number of physical port on FOS switch
        :type port_id: string

        :returns: None
        :rtype: None

        """
        method = "set_vlan"
        self.change_mode(MODE_INTERFACE, port_id)
        LOG.debug(_("FOSSW client received: %s"),
                  self._exec_command("switchport mode access"))
        cmd = self._format_command("switchport access vlan {vlan_id}",
                                   vlan_id=segmentation_id)
        if "VLAN ID not found." in self._exec_command(cmd):
            LOG.exception(_LE("VLAN(%s) does not exist on FOS switch. "
                              "Please check vlan setting."), segmentation_id)
            raise FOSSWClientException(method)

    def clear_vlan(self, segmentaion_id, port_id):
        """Disassociate VLAN from specified physical port on FOS switch.

        :param segmentation_id: id of VLAN to be disassociated
        :type segmentation_id: string
        :param port_id: the number of physical port on FOS switch
        :type port_id: string

        :returns: None
        :rtype: None

        """
        self.change_mode(MODE_INTERFACE, port_id)
        self._exec_command("no switchport access vlan")
        self._exec_command("no switchport mode")

    def get_free_logical_port(self):
        """Get logical port number which not used by any LAG configuration.

        :returns: the number of logical port
        :rtype: string

        """
        cmd = SHOW_PC_BRIEF + " | exclude Dynamic begin 3/"
        res = self._exec_command(cmd)
        return res[res.find('3/'):].split(" ")[0]

    def join_to_lag(self, port, logicalport):
        """Join specified port to LAG configuration.

        :param port: the number of physical port on FOS switch
        :type port: string
        :param logicalport: the number of logical port on FOS switch
        :type logicalport: string

        :returns: None
        :rtype: None

        """
        self.change_mode(MODE_INTERFACE, ifname=logicalport)
        self._exec_command("no port-channel static")

        self.change_mode(MODE_INTERFACE, ifname=port)
        self._exec_command("addport {lo_port}".format(lo_port=logicalport))

    def get_vpcid(self, logicalport="none"):
        """Get VPC ID which is associated with specified logical port.

        If this method is called without logicalport, this method returns a
        VPC ID which is not associated with any logicalport.

        :param logicalport: the number of logicalport
            (optional, default: "none")
        :type logicalport: string

        :returns: the ID of VPC or None.
        :rtype: string
        """
        for i in iter(range(1, MAX_VPC_ID + 1)):
            cmd = 'show vpc {vid} | include "Port channel"'.format(vid=str(i))
            tmp_text = self._exec_command(cmd)
            switch_logicalport = tmp_text[tmp_text.find('. ') + 2:]
            if switch_logicalport == logicalport:
                return str(i)
        if logicalport is "none":
            # NOTE(miyagishi_t): All vpc is already associated to any logical
            # port. Therefore, the FOS switch cannot to define a new vpc.
            LOG.error(_LE("There is no free vpc. All vpc is already "
                          "configured."))
        else:
            # NOTE(miyagishi_t): There is no VPC id which is related to
            # specified logicalport. It maybe the vpc already cleard or
            # originaly not defined on the FOS switch.
            LOG.warning(_LW("A vpc which related to logicalport(%s) on FOS "
                            "switch not found."), logicalport)

    def join_to_vpc(self, logicalport, vpcid):
        """Join a specified logical port to member of VPC.

        :param logicalport: the number of logical port on FOS switch
        :type logicalport: string
        :param vpcid: id of VPC
        :type vpcid: string

        :returns: None
        :rtype: None

        """
        self.changemode(MODE_INTERFACE, ifname=logicalport)
        self._exec_commmand("vpc {vpcid}".format(vpcid=vpcid))

    def get_peerlink_partner(self):
        """Get peerlink partner switch's IP address.

        :returns: IP address of peerlink partner
        :rtype: string

        """
        ret = self._exec_command('show vpc peer-keepalive | include "Peer '
                                 'IP address"')
        return ret[ret.find('. ') + 2:]

    def get_lag_port(self, portname):
        """Get logicalport from FOS switch.

        :param portname: the number of physical port on FOS switch
        :type portname: string

        :returns: the number of logical port which associated with specified
                  physical port
        :rtype: string

        """
        key_charactors = [',', ' ']
        for key in key_charactors:
            cmd = SHOW_PC_BRIEF + " | include " + '"' + portname + key + '"'
            res = self._exec_command(cmd)
            if portname in res:
                return res[:res.find(" ")]

    def get_switch_mac(self):
        """Get MAC address of FOS switch.

        :returns: MAC address of FOS switch
        :rtype: string

        """
        return self._exec_command(
            'show hardware eeprom | include "Base MAC Address"')

    def leave_from_lag(self, port, logicalport):
        """Leave a specified port from LAG configuration.

        :param port: the number of physical port on FOS switch
        :type port: string
        :param logicalport: the number of logical port on FOS switch
        :type logicalport: string

        :returns: None
        :rtype: None

        """
        self.change_mode(MODE_INTERFACE, ifname=port)
        res = self._exec_command(
            "deleteport {lo_port}".format(lo_port=logicalport))

        self.change_mode(MODE_INTERFACE, ifname=logicalport)
        self._exec_command("port-channel static")

        if "is not a member of port-channel" in res:
            LOG.warning(_LW("specified port(%(port)s) has already removed "
                            "from logical port(%(log_po)s)"),
                        {"port": port, "log_po": logicalport})

    def leave_from_vpc(self, logicalport, vpcid):
        """Leave a specified logical port from member of VPC.

        :param logicalport: the number of logical port on FOS switch
        :type logicalport: string
        :param vpcid: id of VPC
        :type vpcid: string

        :returns: None
        :rtype: None

        """
        self.change_mode(MODE_INTERFACE, ifname=logicalport)
        res = self._exec_command("no vpc {vpcid}".format(vpcid=vpcid))
        if "Failed to remove" in res:
            LOG.warning(_LW("specified logical port(%(log_po)s) has already "
                            "removed from VPC(%(vpc)s)."),
                        {"log_po": logicalport, "vpc": vpcid})

    def save_running_config(self):
        """Save running config of FOS switch.

        :returns: None
        :rtypes: None

        """
        self._exec_command("copy system:running-config nvram:startup-config")
        self._exec_command("y")
        self._exec_command("")

    def change_mode(self, mode, ifname=None):
        """Change CLI mode of FOS switch.

        :param mode: the name of the mode to be moved.
        :type mode: string
        :param ifname: the number of physical port on FOS switch
            (optional, default: None)
        :type ifname: string

        :returns: None
        :rtype: None

        """
        # Move to Privileged EXEC mode.
        prompt = self._exec_command(END)
        if ") >" in prompt:
            prompt = self._exec_command(ENABLE)

        if mode == MODE_GLOBAL:
            prompt = self._exec_command(MODE_GLOBAL)
        if mode == MODE_VLAN:
            prompt = self._exec_command(MODE_VLAN)
        if mode == MODE_INTERFACE:
            prompt = self._exec_command(MODE_GLOBAL)
            command = self._format_command(MODE_INTERFACE + " {ifname}",
                                           ifname=ifname)
            prompt = self._exec_command(command)


class FOSSWClientException(Exception):

    def __init__(self, method):
        self.message = _("FOSSW client caught an error at %s") % method

    def __str__(self):
        return(self.message)



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#


"""Implementation of Fujitsu ML2 Mechanism driver for ML2 Plugin."""

from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import importutils

from networking_fujitsu._i18n import _LE
from networking_fujitsu._i18n import _LI
from networking_fujitsu._i18n import _LW
from networking_fujitsu.ml2.common import utils as fj_util

from neutron_lib import constants as const

from neutron.extensions import portbindings
from neutron.plugins.ml2.common import exceptions as ml2_exc
from neutron.plugins.ml2 import driver_api


LOG = logging.getLogger(__name__)
FUJITSU_DRIVER = 'networking_fujitsu.ml2.'
FOSSW_VLAN_DRIVER = FUJITSU_DRIVER + 'fossw.fossw_vlandriver.FOSSWVlanDriver'
FOSSW_VXLAN_DRIVER = FUJITSU_DRIVER + \
    'fossw.fossw_vxlandriver.FOSSWVxlanDriver'

_SUPPORTED_NET_TYPES = ['vlan', 'vxlan']

ML2_FUJITSU_GROUP = "fujitsu_fossw"
ML2_FUJITSU = [
    cfg.ListOpt(
        'fossw_ips', default='[]',
        help=_("List of IP addresses of all FOS switches.")),
    cfg.StrOpt(
        'username', default='',
        help=_("Username of FOS switches.")),
    cfg.StrOpt(
        'password', default='',
        help=_("Password of FOS switches.")),
    cfg.IntOpt(
        'port', default=22,
        help=_("The port number which is used for SSH connection.")),
    cfg.IntOpt(
        'timeout', default=30,
        help=_("The timeout of SSH connection.")),
    cfg.IntOpt(
        'udp_dest_port', default=4789,
        help=_("The port number of VXLAN UDP destionation on the FOS "
               "switch. All VXLANs on the switches use this UDP port as the "
               "UDP destination port in the UDP header when encapsulating.")),
    cfg.IntOpt(
        'ovsdb_vlanid_range_min', default=2,
        help=_("The minimum VLAN ID in the range that is used for binding "
               "VNI and physical port. The range of 78 VLAN IDs (starts from "
               "this value) will be reserved. Default is 2 (VLAN ID from 2 "
               "to 79 will be reserved).")),
    cfg.IntOpt(
        'ovsdb_port', default=6640,
        help=_("The port number which OVSDB server on the FOS switches "
               "lisetn."))]

cfg.CONF.register_opts(ML2_FUJITSU, ML2_FUJITSU_GROUP)


class FOSSWMechanismDriver(driver_api.MechanismDriver):
    """ML2 Mechanism driver for Fujitsu FOS switches.

    This is the upper layer driver class that interfaces to lower layer (CLI)
    below.
    """

    def __init__(self):
        self._vlan_driver = None
        self._vxlan_driver = None
        self._list_switch_info = None
        self.initialize()

    def initialize(self):
        """Initialize of variables needed by this class."""

        self.ips = cfg.CONF.fujitsu_fossw.fossw_ips
        LOG.debug(_("loaded switch IP is: %s"), self.ips)
        if len(self.ips) == 0:
            raise cfg.RequiredOptError(
                'fossw_ips', cfg.OptGroup(ML2_FUJITSU_GROUP))
        self.username = cfg.CONF.fujitsu_fossw.username
        self.password = cfg.CONF.fujitsu_fossw.password
        self._vlan_driver = importutils.import_object(FOSSW_VLAN_DRIVER,
                                                      cfg.CONF)
        self._vxlan_driver = importutils.import_object(FOSSW_VXLAN_DRIVER,
                                                       cfg.CONF)
        self.switches_mac_ip_pair = self._vlan_driver.get_switch_mac_ip_pair(
            self.ips
        )
        LOG.debug(_("Registered FOS switch's IP and MAC pairs: %s"),
                  self.switches_mac_ip_pair)

    @log_helpers.log_method_call
    def create_network_postcommit(self, mech_context):
        """Calls setup process for FOS switch.

        Case1: Create VLAN network
                   Create VLAN with specified VLAN ID.
        Case2: Create VXLAN network
                   Create DCVPN with specified VNI.
        Case3: Otherwise
                   Do nothing.

        :param mech_context: context of network
        :type mech_context: NetworkContext

        :returns: None
        :rtype: None

        """

        network = mech_context.current
        network_id = network['id']
        tenant_id = network['tenant_id']
        provider_type = network['provider:network_type']
        segmentation_id = network['provider:segmentation_id']

        if provider_type == 'vlan' and segmentation_id:
            self.create_network_postcommit_vlan(segmentation_id)
        elif provider_type == 'vxlan' and segmentation_id:
            self.create_network_postcommit_vxlan(network_id, segmentation_id)

        LOG.info(
            _LI("created network (postcommit): network_id=%(network_id)s "
                "tenant_id=%(tenant_id)s"),
            {'network_id': network_id, 'tenant_id': tenant_id})

    @log_helpers.log_method_call
    def create_network_postcommit_vlan(self, vlan_id):
        """Create VLAN to FOS switch.

        :param vlan_id: the ID of VLAN to be created
        :type vlan_id: string

        :returns: None
        :rtype: None

        """
        method = 'create_network_postcommit'
        for ip in self.ips:
            try:
                self._vlan_driver.create_vlan(
                    ip,
                    vlan_id
                )
            except Exception as e:
                LOG.exception(_LE("Fujitsu Mechanism: failed to create vlan "
                                  "switch(%(ip)s. %(err)s"),
                              {'ip': ip, 'err': e})
                raise ml2_exc.MechanismDriverError(method=method)

    @log_helpers.log_method_call
    def create_network_postcommit_vxlan(self, net_uuid, vnid):
        method = 'create_network_postcommit'
        try:
            self._vxlan_driver.create_logical_switch(net_uuid, vnid)
        except Exception as e:
            LOG.exception(_LE("Fujitsu Mechanism: failed to create vxlan "
                              "switch(%(ip)). %(err)s"),
                          {'ip': self.ips, 'err': e})
            raise ml2_exc.MechanismDriverError(method=method)

    @log_helpers.log_method_call
    def delete_network_postcommit(self, mech_context):
        """Calls clean process for FOS switch.

        Case1: Baremetal deploy with VLAN network
                   Delete VLAN definition.
        Case2: Baremetal deploy with VXLAN network
                   Delete DCVPN definition.
        Case3: Otherwise:
                   Do nothing.

        :param mech_context: context of network
        :type mech_context: NetworkContext

        :returns: None
        :rtype: None
        """

        network = mech_context.current
        network_id = network['id']
        tenant_id = network['tenant_id']
        provider_type = network['provider:network_type']
        segmentation_id = network['provider:segmentation_id']

        if (provider_type == 'vlan' and segmentation_id and
           not fj_util.is_baremetal(mech_context.current)):
            self.delete_network_postcommit_vlan(segmentation_id)
        elif provider_type == 'vxlan' and segmentation_id:
            self.delete_network_postcommit_vxlan(network_id)
        else:
            return

        LOG.info(
            _LI("deleted network (postcommit): network_id=%(network_id)s "
                "tenant_id=%(tenant_id)s"),
            {'network_id': network_id, 'tenant_id': tenant_id})

    @log_helpers.log_method_call
    def delete_network_postcommit_vlan(self, vlan_id):
        """Calls clean vlan process for FOS switch.

        :param vlan_id: the ID of VLAN to be deleted
        :type vlan_id: string

        :returns: None
        :rtypes: None

        """
        method = 'delete_network_postcommit'
        for ip in self.ips:
            try:
                self._vlan_driver.delete_vlan(
                    ip,
                    vlan_id
                )
            except Exception as e:
                LOG.exception(
                    _LE("Fujitsu Mechanism: failed to validate "
                        "on switch(%(ip)s). %(err)"),
                    {'ip': ip, 'err': e}
                )
                raise ml2_exc.MechanismDriverError(method=method)

    @log_helpers.log_method_call
    def delete_network_postcommit_vxlan(self, net_uuid):
        method = 'delete_network_postcommit'
        try:
            self._vxlan_driver.delete_logical_switch(net_uuid)
        except Exception as e:
            LOG.exception(
                _LE("Fujitsu Mechanism: failed to validate "
                    "on switch(%(ip)). %(err)"),
                {'ip': self.ips, 'err': e}
            )
            raise ml2_exc.MechanismDriverError(method=method)

    @log_helpers.log_method_call
    def delete_port_postcommit(self, mech_context):
        """Calls cleanup process for C-Fabric.

        Case1: Baremetal deploy with VLAN
                   Clear VLAN/LAG for specified physical port.
        Case2:
                   Clear VXLAN/LAG for specified physical port.
        Case3: Otherwise
                   Do nothing.

        :param mech_context: context of port
        :type mech_context: PortContext

        :returns: None
        :rtype: None
        """

        method = 'delete_port_postcommit'
        port = mech_context.current
        port_id = port['id']
        network_id = port['network_id']
        tenant_id = port['tenant_id']
        segment = mech_context.network.network_segments[0]
        provider_type = segment[driver_api.NETWORK_TYPE]

        if (provider_type == 'vlan' and
           fj_util.is_baremetal(mech_context.current)):
            self.delete_port_postcommit_vlan(mech_context, method)
        elif provider_type == 'vxlan':
            self.delete_port_postcommit_vxlan(mech_context, method)
        else:
            return
        LOG.info(
            _LI("delete port (postcommit): port_id=%(port_id)s "
                "network_id=%(network_id)s tenant_id=%(tenant_id)s"),
            {'port_id': port_id,
             'network_id': network_id, 'tenant_id': tenant_id})

    @log_helpers.log_method_call
    def delete_port_postcommit_vlan(self, context, method):
        """Clear VLAN from specified physical port on switch.

        :param context: context of port
        :type context: PortContext
        :param method: the name of method
        :type method: string

        :returns: None
        :rtype: None

         """
        if validate_baremetal_deploy(context):
            params = self.get_physical_net_params(context)
            try:
                self.clear_vlan(params)
            except Exception:
                LOG.exception(
                    _LE("Failed to clear vlan(%s)."),
                    params['vlan_id']
                )
                raise ml2_exc.MechanismDriverError(method=method)

    @log_helpers.log_method_call
    def update_port_postcommit(self, context):
        """Update VXLAN from specified physical port on switch."""
        port_context = context.current
        if fj_util.is_baremetal(port_context):
            return
        network = context.network
        vif_type = port_context['binding:vif_type']
        network_type = network.network_segments[0][driver_api.NETWORK_TYPE]
        if network_type != 'vxlan' or vif_type == 'unbound':
            return
        # currently supports only one segment per network
        vni = network.network_segments[0][driver_api.SEGMENTATION_ID]
        try:
            self._vxlan_driver.update_physical_port(
                vni, [], port_context, self.switches_mac_ip_pair)
        except Exception:
            LOG.exception(_LE("Failed to create vxlan(%s)."), vni)
            raise ml2_exc.MechanismDriverError(method='create_port_postcommit')

    @log_helpers.log_method_call
    def delete_port_postcommit_vxlan(self, context, method):
        """Clear VXLAN from specified physical port on switch."""
        port_context = context.current
        network = context.network
        network_type = network.network_segments[0][driver_api.NETWORK_TYPE]
        vif_type = port_context['binding:vif_type']
        if vif_type == 'unbound' or network_type == 'flat':
            return
        # currently supports only one segment per network
        segment = context.network.network_segments[0]
        vni = segment[driver_api.SEGMENTATION_ID]
        lli = fj_util.get_physical_connectivity(port_context)
        lag = fj_util.is_lag(lli)
        target = 'reset_physical_port_with_lag' \
            if lag else 'reset_physical_port'
        try:
            reset_method = getattr(self._vxlan_driver, target)
        except AttributeError:
            LOG.exception(_LE("Unexpected error happend."))
            raise ml2_exc.MechanismDriverError(
                method='delete_port_postcommit_vxlan')
        try:
            reset_method(lli, port_context, self.switches_mac_ip_pair)
        except Exception:
            LOG.exception(_LE("Failed to clear vxlan(%s)."), vni)
            raise ml2_exc.MechanismDriverError(method=method)

    @log_helpers.log_method_call
    def setup_vlan(self, params):
        """Provision VLAN with specified port(s).

        This method will select driver's method.
        Case1: param['lag'] is True
            This method calls 'setup_vlan_with_lag' and clears VLAN and LAG.
        Case2: param['lag'] is False
            This method calls 'setup_vlan' and setup only VLAN.

        :param params: a dictionary of the return value for
                        get_physical_net_params
        :type params: dictionary

        :returns: None
        :rtype: None
        """

        target = 'setup_vlan_with_lag' if params['lag'] else 'setup_vlan'
        try:
            setup_method = getattr(self._vlan_driver, target)
        except AttributeError:
            LOG.exception(_LE("Unexpected error happend."))
            raise ml2_exc.MechanismDriverError(method='setup_vlan')
        try:
            # This plugin supposes 1 C-Fabric(fabric_id) management.
            # Therefore, not to identify target IP address by using
            # switch_info(mac_address).
            LOG.info(
                _LI("call %(target)s.  params: %(params)s"),
                {'target': target, 'params': params}
            )
            setup_method(
                params['vlan_id'],
                params['local_link_info'],
                self.switches_mac_ip_pair,
            )
        except Exception:
            LOG.exception(_LE("Fujitsu Mechanism: "
                              "failed to setup vlan %s"), params['vlan_id'])
            raise ml2_exc.MechanismDriverError(method='setup_vlan')

    @log_helpers.log_method_call
    def clear_vlan(self, params):
        """Clear VLAN with specified port(s).

        This method will select driver's method.
        Case1: param['lag'] is True
            This method calls 'clear_vlan_with_lag' and clears VLAN and LAG.
        Case2: param['lag'] is False
            This method calls 'clear_vlan' and clears only VLAN.

        :param params: a dictionary of the return value for
                       get_physical_net_params
        :type params: dictionary
        :returns: None
        :rtype: None
        """

        target = 'clear_vlan_with_lag'
        call_target = target if params['lag'] else 'clear_vlan'
        try:
            clear_method = getattr(self._vlan_driver, call_target)
        except AttributeError:
            LOG.exception(_LE("Unexpected error happend."))
            raise ml2_exc.MechanismDriverError(method="clear_vlan")

        try:
            # This plugin supposes 1 C-Fabric(fabric_id) management.
            # Therefore, not to identify target IP address by using
            # switch_info(mac_address).
            LOG.info(
                _LI("call %(target)s.  params: %(params)s"),
                {'target': call_target, 'params': params}
            )
            clear_method(
                params['vlan_id'],
                params['local_link_info'],
                self.switches_mac_ip_pair,
            )
        except Exception:
            LOG.exception(_LE("Fujitsu Mechanism: "
                              "failed to clear vlan %s"), params['vlan_id'])
            raise ml2_exc.MechanismDriverError(method="clear_vlan")

    @log_helpers.log_method_call
    def get_physical_net_params(self, mech_context):
        """Validate physical network parameters for baremetal deployment.

        Validates network & port params and returns dictionary.
        'local_link_information' is a dictionary from Ironic-port.  This value
        includes as follows:
            'switch_id': A string of switch's MAC address
                         This value is equal to 'chassis_id' from LLDP TLV.
            'port_id': A string of switch interface name.
                         This value is equal to 'port_id' from LLDP TLV.
            'switch_info': A string of switch name.
                         This value is equal to 'system_name' from LLDP TLV.

        :param mech_context: a Context instance
        :type mech_context: PortContext

        :returns: A dictionary parameters for baremetal deploy
        :rtype: dictionary
        """

        port = mech_context.current
        # currently supports only one segment per network
        segment = mech_context.network.network_segments[0]
        vlan_id = segment[driver_api.SEGMENTATION_ID]
        local_link_info = fj_util.get_physical_connectivity(port)
        return {
            "local_link_info": local_link_info,
            "vlan_id": vlan_id,
            "mac": port['mac_address'],
            "lag": fj_util.is_lag(local_link_info)
        }

    @log_helpers.log_method_call
    def bind_port(self, context):

        port = context.current
        network = context.network
        vnic_type = port['binding:vnic_type']
        LOG.debug("Attempting to bind port %(port)s with vnic_type "
                  "%(vnic_type)s on network %(network)s",
                  {'port': port['id'], 'vnic_type': vnic_type,
                   'network': context.network.current['id']})
        provider_type = network.network_segments[0][driver_api.NETWORK_TYPE]
        segments = context.segments_to_bind
        if provider_type == 'vlan' and validate_baremetal_deploy(context):
            self.setup_vlan(self.get_physical_net_params(context))
        if provider_type == 'vxlan':
            self.bind_port_vxlan(context)
        context.set_binding(segments[0][driver_api.ID],
                            portbindings.VIF_TYPE_OTHER, {},
                            status=const.PORT_STATUS_ACTIVE)

    @log_helpers.log_method_call
    def bind_port_vxlan(self, context):
        """Update VXLAN from specified physical port on switch."""
        # currently supports only one segment per network
        port_context = context.current
        segment = context.network.network_segments[0]
        vnid = segment[driver_api.SEGMENTATION_ID]
        lli = fj_util.get_physical_connectivity(port_context)
        lag = fj_util.is_lag(lli)
        target = 'update_physical_port_with_lag' \
            if lag else 'update_physical_port'
        try:
            update_method = getattr(self._vxlan_driver, target)
        except AttributeError:
            LOG.exception(_LE("Unexpected error happend."))
            raise ml2_exc.MechanismDriverError(method='bind_port_vxlan')

        try:
            update_method(vnid, lli, port_context, self.switches_mac_ip_pair)
        except Exception:
            LOG.exception(_LE("Failed to create vxlan(%s)."), vnid)
            raise ml2_exc.MechanismDriverError(method='bind_port_vxlan')


def is_supported(network):
    """Validate network parameter(network_type and segmentation_id).

    :param network: a network object
    :type network: NetworkContext

    :returns: True if network_type is 'VLAN' and segmentation_id is included
              otherwise False
    :rtype: boolean
    """

    segment = network.network_segments[0]
    seg_id = segment[driver_api.SEGMENTATION_ID]
    net_type = segment[driver_api.NETWORK_TYPE]
    if (net_type in _SUPPORTED_NET_TYPES and seg_id):
        return True
    LOG.warning(_LW("%s is not supported. Skip it."), net_type)
    return False


def validate_baremetal_deploy(mech_context):
    """Validate baremetal deploy.

    :param mech_context: a context object
    :type mech_context: PortContext

    :returns: True if enable to baremetal deploy otherwise False
    :rtype: boolean
    """

    port = mech_context.current
    network = mech_context.network
    if (fj_util.is_baremetal(port) and
       is_supported(network) and fj_util.get_physical_connectivity(port)):
        return True
    return False



# Copyright 2017 FUJITSU LIMITED
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from oslo_config import cfg
from oslo_log import log as logging

from networking_fujitsu._i18n import _LI
from networking_fujitsu._i18n import _LW
from networking_fujitsu.ml2.common.ovsdb import ovsdb_writer
from networking_fujitsu.ml2.common import type_vxlan
from networking_fujitsu.ml2.fossw import client
from networking_fujitsu.ml2.fossw import fossw_vlandriver

from neutron.common import utils

LOG = logging.getLogger(__name__)
_LOCK_NAME = 'fujitsu_fossw'


class FOSSWVxlanDriver(object):
    """Fossw CLI interface driver for Neutron network.

    Handles life-cycle management of Neutron network (OVSDB on FOSSW)
    """

    def __init__(self, conf=None):
        if conf:
            self._conf = conf
        else:
            self._conf = cfg.CONF
        self.ovsdb_port = self._conf.fujitsu_fossw.ovsdb_port
        self.udp_dest_port = self._conf.fujitsu_fossw.udp_dest_port
        self.ovsdb_vlanid_range_min = (
            self._conf.fujitsu_fossw.ovsdb_vlanid_range_min)
        self.fossw_ips = self._conf.fujitsu_fossw.fossw_ips

        self.client = client.FOSSWClient(self._conf)
        self.type_vxlan = type_vxlan.TypeVxlan()
        self.vxlan_endpoint_ips = []
        self.initialize()

    def initialize(self):
        """Initialize update

        Both update Neutron DB with switch VTEPs, and update FOS switches
        with host & other switches' VTEPs.
        """

        if "vxlan" in self._conf.ml2.type_drivers:
            self._update_neutron_db()
            self.vxlan_endpoint_ips = [vxlan_endpoint['ip_address']
                                       for vxlan_endpoint in
                                       self.type_vxlan.get_endpoints()]

    def _save_all_fossw(self):
        """Save running-config to startup-config in all FOS switches."""
        for fossw_ip in self.fossw_ips:
            self.client.connect(fossw_ip)
            self.client.save_running_config()
            self.client.disconnect()

    def _update_neutron_db(self):
        """Update Neutron DB ml2_vxlan_endpoints table

        Update with VTEPs of all FOS switches.
        :returns: None
        """

        for target_ip in self.fossw_ips:
            ovsdb_client = ovsdb_writer.OVSDBWriter(target_ip, self.ovsdb_port)
            sw_ep_ip, sw_ep_host = ovsdb_client.get_sw_ep_info()
            if not sw_ep_host:
                sw_ep_host = "FOSSW"
            if sw_ep_ip:
                neutron_db = self.type_vxlan.get_endpoint_by_ip(
                    sw_ep_ip)
                if not neutron_db:
                    LOG.info(_LI("Add FOS switch endpoint information (ip = "
                                 "%(ip)s, hostname = %(name)s, udp_dest_port "
                                 "= %(port)s) to ml2_vxlan_endpoints table in "
                                 "Neutron DB."), {'ip': sw_ep_ip,
                                                  'name': sw_ep_host,
                                                  'port': self.udp_dest_port})
                    self.type_vxlan.add_endpoint(sw_ep_ip, sw_ep_host,
                                                 self.udp_dest_port)
                elif (neutron_db['host'] != sw_ep_host or
                      neutron_db['udp_port'] != self.udp_dest_port):
                    LOG.warning(_LW("Existed endpoint information with same "
                                    "IP address (%(ip)s) has been updated "
                                    "with hostname = %(host)s, udp_dest_port "
                                    "= %(port)s."),
                                {'ip': sw_ep_ip, 'host': sw_ep_host,
                                 'port': self.udp_dest_port})
                    self.type_vxlan.delete_endpoint(sw_ep_ip)
                    self.type_vxlan.add_endpoint(sw_ep_ip, sw_ep_host,
                                                 self.udp_dest_port)
                else:
                    LOG.info(_LI("FOS switch endpoint information (ip = "
                                 "%(ip)s, hostname = %(name)s, udp_dest_port "
                                 "= %(port)s) has alreadly been added to "
                                 "ml2_vxlan_endpoints table in Neutron DB. "
                                 "Do nothing."),
                             {'ip': sw_ep_ip, 'name': sw_ep_host,
                              'port': self.udp_dest_port})
            else:
                LOG.warning(_LW("Unable to get endpoint information from "
                                "switch (IP = %s). Skip."), target_ip)

    @utils.synchronized(_LOCK_NAME, external=True)
    def create_logical_switch(self, net_uuid, vnid):
        """Create a row in Logical_Switch table in FOS switch OVSDB.

        We consider net_uuid is always unique, and both vnid and net_uuid
        must be immutable. So there is not a update case.
        :param net_uuid: The uuid of Neutron network.
        :type net_uuid: string
        :param vnid: The segment ID of Neutron network.
        :type vnid: integer

        :returns: None
        """

        logical_switch_name = net_uuid.replace("-", "")
        for switch_ip in self.fossw_ips:
            ovsdb_client = ovsdb_writer.OVSDBWriter(switch_ip,
                                                    self.ovsdb_port)
            ovsdb_client.insert_logical_switch(vnid, logical_switch_name)
        self._save_all_fossw()

    @utils.synchronized(_LOCK_NAME, external=True)
    def delete_logical_switch(self, net_uuid):
        """Delete row in Logical_Switch table in FOS switch OVSDB.

        We only consider deleting the existing Logical_Switch ROW.
        Also note that even it does not exit, the jsonrpc returns no error.
        :param net_uuid: The uuid of Neutron network.
        :type: string

        :returns: None
        """

        logical_switch_name = net_uuid.replace("-", "")
        for switch_ip in self.fossw_ips:
            ovsdb_client = ovsdb_writer.OVSDBWriter(switch_ip,
                                                    self.ovsdb_port)
            logical_switch_uuid = ovsdb_client.get_logical_switch_uuid(
                logical_switch_name)
            if logical_switch_uuid:
                ovsdb_client.delete_logical_switch(logical_switch_uuid)
        self._save_all_fossw()

    def update_physical_port(self, vnid, lli, port_context, ip_mac_pairs):
        """Update Physical_Port table in FOS switch OVSDB.

        There are 3 cases about port operation
        Case 1:
        Port in FOS switch need to get bind. In this case, Ucast_Macs_Local
        table and Physical_locator table also need to insert a new row or
        update at the same time.
        Case 2:
        Another FOS switch has updated its port. In this case, other FOS
        switches need to sync the information to Ucast_Macs_Remote table and
        Physical_locator table.
        Case 3:
        Nova Compute Node has created a new VM or something others occurred
        around Neutron network. In this case, all FOS switches should do the
        same thing as in Case 2.

        :param vnid: The segment ID of Neutron network which the port belongs
                     to.
        :type vnid: integer
        :param lli: The local_link_information of the port.
                    If it is a VM port, then [{}] should be given.
        :type lli: list
        :param port_context: Context of the port.
        :type port_context: dictionary
        :param ip_mac_pairs: List of MAC(key) - IP(value) pairs of all FOS
                             switches.
        :type ip_mac_pairs: dictionary

        :returns: None
        """

        if lli:
            target_name = lli[0]['switch_info']
            target_ip = ip_mac_pairs[lli[0]['switch_id']]
            port_id = lli[0]['port_id']
        else:
            target_name = port_context['binding:host_id']
            target_ip = ""
            # In this case, it will be a VM port, and we do not need to
            # care about port_id.
        target_tunnel_ip = self.type_vxlan.db_get_endpoint_ip_by_host(
            target_name)
        net_uuid = port_context["network_id"]
        port_mac = port_context["mac_address"]
        fixed_ips = port_context["fixed_ips"]
        port_ips = []
        for fixed_ip in fixed_ips:
            port_ips.append(fixed_ip["ip_address"])

        logical_switch_name = net_uuid.replace("-", "")
        if target_ip in self.fossw_ips:
            # Update Physical_Port table first.
            ovsdb_client = ovsdb_writer.OVSDBWriter(target_ip, self.ovsdb_port)
            bind_ls_uuid = ovsdb_client.get_logical_switch_uuid(
                logical_switch_name)
            binding_vid = ovsdb_client.get_binding_vid(bind_ls_uuid)
            bind_vid = binding_vid if binding_vid else (
                int(port_id[2:]) + self.ovsdb_vlanid_range_min - 1)

            ovsdb_client.update_physical_port(port_id, bind_vid, bind_ls_uuid)
            # After Physical_Port table has been updated, update
            # Ucast_Macs_Local table.
            # If any garbage exist, remove them first.
            if ovsdb_client.get_ucast_macs_local(port_mac):
                ovsdb_client.delete_ucast_macs_local(port_mac)
            # Then add what we want to the Ucast_Macs_Local table.
            locator_ip_local, sw_ep_host = ovsdb_client.get_sw_ep_info()
            locator_uuid_local = ovsdb_client.get_physical_locator_uuid(
                locator_ip_local)
            # We do not care about port IP address. We only forward using port
            # MAC address.
            if locator_uuid_local:
                ovsdb_client.insert_ucast_macs_local(
                    bind_ls_uuid, locator_uuid_local, port_mac)
            else:
                ovsdb_client.insert_ucast_macs_local_and_locator(
                    bind_ls_uuid, locator_ip_local, port_mac)

        # At last Ucast_Macs_Remote table of other switches.
        self._update_ucast_macs_remote(target_ip, logical_switch_name,
                                       port_mac, target_tunnel_ip, port_ips)
        self._save_all_fossw()

    def _update_ucast_macs_remote(self, target_ip, logical_switch_name,
                                  port_mac, target_tunnel_ip, port_ips):
        """Update Ucast_Macs_Remote table in all FOS switches OVSDB."""
        for fossw_ip in self.fossw_ips:
            if fossw_ip != target_ip:
                ovsdb_client = ovsdb_writer.OVSDBWriter(
                    fossw_ip, self.ovsdb_port)
                ls_uuid = ovsdb_client.get_logical_switch_uuid(
                    logical_switch_name)
                if ovsdb_client.get_ucast_macs_remote(port_mac):
                    ovsdb_client.delete_ucast_macs_remote(port_mac)
                locator_uuid = ovsdb_client.get_physical_locator_uuid(
                    target_tunnel_ip)
                if locator_uuid:
                    ovsdb_client.insert_ucast_macs_remote(
                        ls_uuid, port_mac, port_ips, locator_uuid)
                else:
                    ovsdb_client.insert_ucast_macs_remote_and_locator(
                        ls_uuid, port_mac, port_ips, target_tunnel_ip)

    def reset_physical_port(self, lli, port_context, ip_mac_pairs):
        """Remove setting of raw of Physical_Port table in FOS switch OVSDB.

        ROWs with the same MAC address in Ucast_Macs_Local table and
        Ucast_Macs_Remote table are the target to extinct.

        :param lli: The local_link_information of the port.
        :type lli: list
        :param port_context: Context of the port.
        :type port_context: dictionary
        :param ip_mac_pairs: List of MAC(key) - IP(value) pairs of all FOS
                             switches.
        :type ip_mac_pairs: dictionary

        :returns: None
        """

        if lli:
            target_ip = ip_mac_pairs[lli[0]['switch_id']]
            port_id = lli[0]['port_id']
        else:
            target_ip = ""
            # In this case, it will be a VM port, and we do not need to
            # care about port_id.

        port_mac = port_context["mac_address"]

        if target_ip in self.fossw_ips:
            ovsdb_client = ovsdb_writer.OVSDBWriter(target_ip,
                                                    self.ovsdb_port)
            ovsdb_client.reset_physical_port(port_id)
        for fossw_ip in self.fossw_ips:
            # All Ucast_Macs_Remote and Ucast_Macs_Local tables in all
            # FOS switches will delete the ROW with the MAC address.
            ovsdb_client = ovsdb_writer.OVSDBWriter(fossw_ip,
                                                    self.ovsdb_port)
            ovsdb_client.delete_ucast_macs_local(port_mac)
            ovsdb_client.delete_ucast_macs_remote(port_mac)
        self._save_all_fossw()

    @utils.synchronized(_LOCK_NAME, external=True)
    def update_physical_port_with_lag(self, vnid, llis, port_context,
                                      ip_mac_pairs):
        """Update Physical_Port table in FOS switch OVSDB.

        (lag case)
        :param vnid: The segment ID of Neutron network which the port belongs
                     to.
        :type vnid: integer
        :param llis: `local_link_information' of 'binding:profile' for the port
                     multiple physical ports(dict) should be given.
        :type llis: list
        :param port_context: Context of the port.
        :type port_context: dictionary
        :param ip_mac_pairs: List of MAC(key) - IP(value) pairs of all FOS
                             switches.
        :type ip_mac_pairs: dictionary

        :returns: None
        """

        vlan_id = 1
        fossw_vlandriver.FOSSWVlanDriver().setup_vlan_with_lag(
            vlan_id, llis, ip_mac_pairs)

        # setup ovsdb for each physical port
        for lli in llis:
            self.update_physical_port(vnid, lli, self.fossw_ips,
                                      port_context, ip_mac_pairs)

    @utils.synchronized(_LOCK_NAME, external=True)
    def reset_physical_port_with_lag(self, llis, port_context, ip_mac_pairs):
        """Remove setting of raw of Physical_Port table in FOS switch OVSDB.

        (lag case)
        :param llis: `local_link_information' of 'binding:profile' for the port
                     multiple physical ports(dict) should be given.
        :type llis: list
        :param port_context: Context of the port.
        :type port_context: dictionary
        :param ip_mac_pairs: List of MAC(key) - IP(value) pairs of all FOS
                             switches.
        :type ip_mac_pairs: dictionary

        :returns: None
        """

        vlan_id = 1
        fossw_vlandriver.FOSSWVlanDriver().clear_vlan_with_lag(
            vlan_id, llis, ip_mac_pairs)

        # Clear lag setting
        for lli in llis:
            self.reset_physical_port(lli, self.fossw_ips,
                                     port_context, ip_mac_pairs)
